
;;; b5 caching evaluator

;; smashing.

(define (cons! pair head tail)
  (set-car! pair head)
  (set-cdr! pair tail)
  pair)

(define (smash! dst src)
  (cons! dst (car src) (cdr src)))

(define (smash-if-possible! dst src)
  (cond ((and (pair? dst) (pair? src)) (smash! dst src))
	((pair? dst) (cons! dst 'res (list src "")))
	(else src)))

;; evaluation.

(define (expr-eval-ident ident)
  (expr-eval (hash-table-ref *macros* ident (symbol->string ident))))

(define (expr-eval-sequence seq)
  (smash-if-possible! seq       ; this breaks tail recursion
    (let loop ((clos (cadr seq)) (seq (cddr seq)))
      (if (null? seq) (expr-eval clos)
	(loop (expr-apply clos (car seq)) (cdr seq))))))

(define *expand-parenthese-contents* #t)

(define (expr-eval expr)
  (case (expr-type expr)
    ((ident) (expr-eval-ident expr))
    ((paren) (if *expand-parenthese-contents*
	       (cons! expr 'paren (expr-eval (cdr expr)))
	       expr))
    ((seq) (expr-eval-sequence expr))
    ((clos) (closure-expand expr))
    ((lit nlit empty res pacl) expr)
    (else (error "expr-eval: unknown expression: " expr))))

;; application.

(define (expr-apply-closure clos expr)
  (closure-expand (closure-add-argument clos expr)))

(define (expr-apply-possibly-applicable-closure pacl expr)
  (expr-apply-closure (cadr pacl) (expr-apply (caddr pacl) expr)))

(define (expr-apply-result res expr)
  (let retry ((res-prefix (cadr res))
	      (res2 (expr-apply (caddr res) expr)))
    (if (expr-type? res2 'res)
      (retry (expr-combine res-prefix (cadr res2)) (caddr res2))
      (list 'res res-prefix res2))))

(define (expr-apply-literal nlit expr)
  (if (expr-type? expr 'nlit)
    (string-append nlit expr)
    (list 'res nlit expr)))

(define (expr-apply expr arg)
  (if (expr-empty? arg) expr
    (case (expr-type expr)
      ((nlit) (expr-apply-literal expr arg))
      ((clos) (expr-apply-closure expr arg))
      ((res) (expr-apply-result expr arg))
      ((pacl) (expr-apply-possibly-applicable-closure expr arg))
      ((ident) (expr-apply (expr-eval-ident expr) arg))
      ((seq) (expr-apply (expr-eval-sequence expr) arg))
      ((empty) arg)
      ((paren lit) (list 'res (expr-eval expr) arg))
      (else (error "expr-apply: unknown expression: " expr)))))

;; closures.

(define (closure-add-argument clos arg)
  (cond ((expr-type? arg 'paren)
	 `(clos ,(- (cadr clos) 1)
		,(cons arg (caddr clos)) . ,(cdddr clos)))
	((expr-literal? arg) `(res ,clos ,arg))
	(else `(pacl ,clos ,arg))))

(define (closure-expand clos)
  (if (or (not (expr-type? clos 'clos)) (> (cadr clos) 0)) clos
    (expr-eval
      (smash-if-possible! clos
	(substitute-arguments
	  (list->vector (reverse (expr-closure-arguments clos)))
	  (expr-closure-body clos))))))

(define (expr-strip-paren expr)
  (if (expr-type? expr 'paren) (cdr expr) expr))

(define (substitute-arguments args body)
  (cond ((and (pair? body) (not (eq? 'clos (car body))))
	 (map (lambda (b) (substitute-arguments args b)) body))
	((integer? body) (expr-strip-paren (vector-ref args body)))
	(else body)))

