
;;; b5 syntax parsers.

;; primitive elements.

(define b5-comment
  (parser-seq-combine
    (lambda (beg rest) "")
    (parser-seq-literal (string->list "///"))
    (parser-chain-endby (lambda (v1 v2) #t)
      (parser-item-literal #\newline) (parser-item char?))))

(define b5-ws 
  (parser-seq-combine
    (lambda (ls) (list->string ls))
    (parser-span1 char-whitespace?)))

(define (b5-separator litws?)
  (parser-bind
    (parser-chain1 string-append "" (parser-alt b5-ws b5-comment))
    (lambda (whitespace) (parser-return (if litws? whitespace "")))))

(define b5-relax (parser-alt (b5-separator #t) (parser-return "")))

(define b5-ident
  (parser-bind
    (parser-span1 (lambda (c) (and (not (char-whitespace? c))
				   (not (memv c '(#\( #\) #\; #\< #\>))))))
    (lambda (ls)
      (parser-return (string->symbol (list->string ls))))))

(define b5-literal
  (parser-seq-combine
    (lambda (beg rest) (list 'lit (list->string rest)))
    (parser-seq-literal (string->list "<<<"))
    (parser-chain-endby cons
      (parser-bind (parser-seq-literal (string->list ">>>"))
		   (lambda (ign) (parser-return '())))
      (parser-item char?))))

;; expressions.

(define (b5-paren litws?)
  (parser-bind
    (parser-item-literal #\()
    (lambda (ign)
      (parser-bind
	(parser-chain-endby cons
          (parser-bind (parser-item-literal #\))
		       (lambda (ign) (parser-return '())))
	  (b5-element litws?))
	(lambda (seq) (parser-return `(paren seq . ,seq)))))))

(define (b5-element litws?)
  (parser-alt (b5-separator litws?) b5-literal (b5-paren litws?) b5-ident))

(define b5-end 
  (parser-bind
    (parser-seq-literal (string->list ";;;"))
    (lambda (ign) (parser-return '()))))

(define (b5-expression litws?)
  (parser-bind
    (parser-chain-endby cons b5-end (b5-element litws?))
    (lambda (seq) (parser-return (cons 'seq seq)))))

(define b5-data
  (parser-bind
    (parser-chain cons '()
      (parser-alt
	(b5-element #t)
	(parser-bind (parser-item char?)
		     (lambda (c) (parser-return (string c))))))
    (lambda (seq) (parser-return (cons 'seq seq)))))

;; statements.

(define b5-formals
  (parser-seq-combine
    (lambda (formals crap) formals)
    (parser-chain-endby cons 
      (parser-bind (parser-seq b5-relax (parser-seq-literal '(#\= #\>)))
		   (lambda (ign) (parser-return '())))
      (parser-bind (b5-separator #f) (lambda (ign) b5-ident)))
    b5-relax))

(define b5-macro-statement
  (parser-seq-combine
    (lambda (beg crap name formals body)
      (add-macro name formals body) "")
    (parser-seq-literal (string->list "mac"))
    b5-relax
    b5-ident
    b5-formals
    (b5-expression #f)))

(define b5-data-statement
  (parser-seq-combine
    (lambda (beg crap contents) contents)
    (parser-seq-literal (string->list "data"))
    b5-relax
    (b5-expression #t)))

(define b5-filename
  (parser-alt
    b5-literal
    (parser-bind b5-ident
		 (lambda (id) (parser-return (symbol->string id))))))

(define (b5-file-statement action keyword)
  (parser-seq-combine
    (lambda (beg crap filename end) (action filename) "")
    (parser-seq-literal (string->list keyword))
    b5-relax
    b5-filename
    b5-end))

(define b5-include-statement (b5-file-statement add-code-file "include"))
(define b5-input-statement (b5-file-statement add-data-file "input"))
(define b5-output-statement (b5-file-statement add-output-stream "output"))
(define b5-stream-statement (b5-file-statement add-stream "stream"))

(define b5-statement
  (parser-alt b5-macro-statement b5-data-statement b5-stream-statement
	      b5-input-statement b5-output-statement b5-include-statement))

(define b5-program
  (parser-chain-endby expr-combine
    (parser-seq-combine
      (lambda (crap1 crap2) "")
      b5-relax parser-end-of-stream)
    (parser-bind b5-relax (lambda (ign) b5-statement))))

